@@ -1,7 +1,6 @@
 # Copyright (c) 2020-2021, Manfred Moitzi
 # License: MIT License
 from typing import (
-    TYPE_CHECKING,
     List,
     Iterable,
     Union,
@@ -49,17 +48,19 @@ from ezdxf.entities import (
     Wipeout,
     MPolygon,
     BoundaryPaths,
-    BoundaryPathType,
-    EdgeType,
+    AbstractBoundaryPath,
+    PolylinePath,
+    EdgePath,
+    LineEdge,
+    ArcEdge,
+    EllipseEdge,
+    SplineEdge,
 )
 from .path import Path
 from .commands import Command
 from . import tools
 from .nesting import group_paths
 
-if TYPE_CHECKING:
-    from ezdxf.entities import TBoundaryPath, PolylinePath, EdgePath
-
 __all__ = [
     "make_path",
     "to_lines",
@@ -253,14 +254,21 @@ def _from_hatch(hatch: Hatch, **kwargs) -> Path:
         from_hatch_boundary_path(boundary, ocs, elevation, offset=offset)
         for boundary in hatch.paths
     ]
+    # looses the boundary path state:
     return tools.to_multi_path(paths)
 
 
 def from_hatch(hatch: Hatch) -> Iterable[Path]:
     """Yield all HATCH boundary paths as separated :class:`Path` objects.
 
+
     .. versionadded:: 0.16
 
+    .. versionchanged:: 17.1
+
+        Attaches the boundary state to each path as
+        :class:`ezdxf.lldxf.const.BoundaryPathState`.
+
     """
     ocs = hatch.ocs()
     elevation = hatch.dxf.elevation.z
@@ -273,28 +281,42 @@ def from_hatch(hatch: Hatch) -> Iterable[Path]:
 
 
 def from_hatch_boundary_path(
-    boundary: "TBoundaryPath",
+    boundary: AbstractBoundaryPath,
     ocs: OCS = None,
     elevation: float = 0,
     offset: Vec3 = NULLVEC,  # ocs offset!
 ) -> "Path":
     """Returns a :class:`Path` object from a :class:`~ezdxf.entities.Hatch`
     polyline- or edge path.
+
+    .. versionchanged:: 17.1
+
+        Attaches the boundary state to each path as
+        :class:`ezdxf.lldxf.const.BoundaryPathState`.
+
     """
-    if boundary.type == BoundaryPathType.EDGE:
-        p = from_hatch_edge_path(boundary, ocs, elevation)  # type: ignore
+    if isinstance(boundary, EdgePath):
+        p = from_hatch_edge_path(boundary, ocs, elevation)
+    elif isinstance(boundary, PolylinePath):
+        p = from_hatch_polyline_path(boundary, ocs, elevation)
     else:
-        p = from_hatch_polyline_path(boundary, ocs, elevation)  # type: ignore
+        raise TypeError(type(boundary))
+
     if offset and ocs is not None:  # only for MPOLYGON
         # assume offset is in OCS
         offset = ocs.to_wcs(offset.replace(z=elevation))
         p = p.transform(Matrix44.translate(offset.x, offset.y, offset.z))
+
+    # attach path type information
+    p.user_data = const.BoundaryPathState.from_flags(
+        boundary.path_type_flags
+    )
     return p
 
 
 def from_hatch_polyline_path(
-    polyline: "PolylinePath", ocs: OCS = None, elevation: float = 0
-) -> "Path":
+    polyline: PolylinePath, ocs: OCS = None, elevation: float = 0
+) -> Path:
     """Returns a :class:`Path` object from a :class:`~ezdxf.entities.Hatch`
     polyline path.
     """
@@ -310,11 +332,11 @@ def from_hatch_polyline_path(
 
 
 def from_hatch_edge_path(
-    edges: "EdgePath",
+    edges: EdgePath,
     ocs: OCS = None,
     elevation: float = 0,
     open_loops: bool = False,
-) -> "Path":
+) -> Path:
     """Returns a :class:`Path` object from a :class:`~ezdxf.entities.Hatch`
     edge path.
 
@@ -322,14 +344,14 @@ def from_hatch_edge_path(
     necessary to override this behavior, by setting `open_loops` to ``True``.
     """
 
-    def line(edge):
+    def line(edge: LineEdge):
         start = wcs(edge.start)
         end = wcs(edge.end)
         segment = Path(start)
         segment.line_to(end)
         return segment
 
-    def arc(edge):
+    def arc(edge: ArcEdge):
         x, y, *_ = edge.center
         # from_arc() requires OCS data:
         # Note: clockwise oriented arcs are converted to counter
@@ -346,7 +368,7 @@ def from_hatch_edge_path(
         tools.add_ellipse(segment, ellipse, reset=True)
         return segment
 
-    def ellipse(edge):
+    def ellipse(edge: EllipseEdge):
         ocs_ellipse = edge.construction_tool()
         # ConstructionEllipse has WCS representation:
         # Note: clockwise oriented ellipses are converted to counter
@@ -364,7 +386,7 @@ def from_hatch_edge_path(
         tools.add_ellipse(segment, ellipse, reset=True)
         return segment
 
-    def spline(edge):
+    def spline(edge: SplineEdge):
         control_points = [wcs(p) for p in edge.control_points]
         if len(control_points) == 0:
             fit_points = [wcs(p) for p in edge.fit_points]
@@ -380,7 +402,7 @@ def from_hatch_edge_path(
         tools.add_spline(segment, bspline, reset=True)
         return segment
 
-    def from_fit_points(edge, fit_points):
+    def from_fit_points(edge: SplineEdge, fit_points):
         tangents = None
         if edge.start_tangent and edge.end_tangent:
             tangents = (
@@ -392,7 +414,7 @@ def from_hatch_edge_path(
             tangents=tangents,
         )
 
-    def from_control_points(edge, control_points):
+    def from_control_points(edge: SplineEdge, control_points):
         return BSpline(
             control_points=control_points,
             order=edge.degree + 1,
@@ -417,16 +439,18 @@ def from_hatch_edge_path(
     loop: Optional[Path] = None
     for edge in edges:
         next_segment: Optional[Path] = None
-        if edge.type == EdgeType.LINE:
+        if isinstance(edge, LineEdge):
             next_segment = line(edge)
-        elif edge.type == EdgeType.ARC:
+        elif isinstance(edge, ArcEdge):
             if abs(edge.radius) > ABS_TOL:
                 next_segment = arc(edge)
-        elif edge.type == EdgeType.ELLIPSE:
+        elif isinstance(edge, EllipseEdge):
             if not Vec2(edge.major_axis).is_null:
                 next_segment = ellipse(edge)
-        elif edge.type == EdgeType.SPLINE:
+        elif isinstance(edge, SplineEdge):
             next_segment = spline(edge)
+        else:
+            raise TypeError(type(edge))
 
         if next_segment is None:
             continue
@@ -479,7 +503,7 @@ def to_lwpolylines(
     distance: float = MAX_DISTANCE,
     segments: int = MIN_SEGMENTS,
     extrusion: "Vertex" = Z_AXIS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[LWPolyline]:
     """Convert the given `paths` into :class:`~ezdxf.entities.LWPolyline`
     entities.
@@ -510,7 +534,7 @@ def to_lwpolylines(
 
     extrusion = Vec3(extrusion)
     reference_point = paths[0].start
-    dxfattribs = dxfattribs or dict()
+    dxfattribs = dict(dxfattribs or {})
     if not Z_AXIS.isclose(extrusion):
         ocs, elevation = _get_ocs(extrusion, reference_point)
         paths = tools.transform_paths_to_ocs(paths, ocs)
@@ -538,7 +562,7 @@ def to_polylines2d(
     distance: float = MAX_DISTANCE,
     segments: int = MIN_SEGMENTS,
     extrusion: "Vertex" = Z_AXIS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[Polyline]:
     """Convert the given `paths` into 2D :class:`~ezdxf.entities.Polyline`
     entities.
@@ -569,7 +593,7 @@ def to_polylines2d(
 
     extrusion = Vec3(extrusion)
     reference_point = paths[0].start
-    dxfattribs = dxfattribs or dict()
+    dxfattribs = dict(dxfattribs or {})
     if not Z_AXIS.isclose(extrusion):
         ocs, elevation = _get_ocs(extrusion, reference_point)
         paths = tools.transform_paths_to_ocs(paths, ocs)
@@ -593,7 +617,7 @@ def to_hatches(
     segments: int = MIN_SEGMENTS,
     g1_tol: float = G1_TOL,
     extrusion: "Vertex" = Z_AXIS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[Hatch]:
     """Convert the given `paths` into :class:`~ezdxf.entities.Hatch` entities.
     Uses LWPOLYLINE paths for boundaries without curves and edge paths, build
@@ -642,7 +666,7 @@ def to_mpolygons(
     distance: float = MAX_DISTANCE,
     segments: int = MIN_SEGMENTS,
     extrusion: "Vertex" = Z_AXIS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[MPolygon]:
     """Convert the given `paths` into :class:`~ezdxf.entities.MPolygon` entities.
     In contrast to HATCH, MPOLYGON supports only polyline boundary paths.
@@ -670,8 +694,8 @@ def to_mpolygons(
     boundary_factory: BoundaryFactory = partial(
         build_poly_path, distance=distance, segments=segments
     )
-    dxfattribs = dxfattribs or dict()
-    dxfattribs.setdefault("fill_color", const.BYLAYER)
+    dxfattribs = dict(dxfattribs or {})
+    dxfattribs.setdefault("fill_color", const.BYLAYER)  # type: ignore
 
     yield from _polygon_converter(
         MPolygon, paths, boundary_factory, extrusion, dxfattribs
@@ -725,7 +749,7 @@ def _polygon_converter(
     paths: Iterable[Path],
     add_boundary: BoundaryFactory,
     extrusion: "Vertex" = Z_AXIS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[TPolygon]:
     if isinstance(paths, Path):
         paths = [paths]
@@ -736,22 +760,22 @@ def _polygon_converter(
 
     extrusion = Vec3(extrusion)
     reference_point = paths[0].start
-    dxfattribs = dxfattribs or dict()
+    _dxfattribs: Dict = dict(dxfattribs or {})
     if not Z_AXIS.isclose(extrusion):
         ocs, elevation = _get_ocs(extrusion, reference_point)
         paths = tools.transform_paths_to_ocs(paths, ocs)
-        dxfattribs["elevation"] = Vec3(0, 0, elevation)
-        dxfattribs["extrusion"] = extrusion
+        _dxfattribs["elevation"] = Vec3(0, 0, elevation)
+        _dxfattribs["extrusion"] = extrusion
     elif reference_point.z != 0:
-        dxfattribs["elevation"] = Vec3(0, 0, reference_point.z)
-    dxfattribs.setdefault("solid_fill", 1)
-    dxfattribs.setdefault("pattern_name", "SOLID")
-    dxfattribs.setdefault("color", const.BYLAYER)
+        _dxfattribs["elevation"] = Vec3(0, 0, reference_point.z)
+    _dxfattribs.setdefault("solid_fill", 1)
+    _dxfattribs.setdefault("pattern_name", "SOLID")
+    _dxfattribs.setdefault("color", const.BYLAYER)
 
     for group in group_paths(tools.single_paths(paths)):
         if len(group) == 0:
             continue
-        polygon = cls.new(dxfattribs=dxfattribs)
+        polygon = cls.new(dxfattribs=_dxfattribs)
         boundaries = polygon.paths
         external = group[0]
         external.close()
@@ -767,7 +791,7 @@ def to_polylines3d(
     *,
     distance: float = MAX_DISTANCE,
     segments: int = MIN_SEGMENTS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[Polyline]:
     """Convert the given `paths` into 3D :class:`~ezdxf.entities.Polyline`
     entities.
@@ -787,7 +811,7 @@ def to_polylines3d(
     if isinstance(paths, Path):
         paths = [paths]
 
-    dxfattribs = dxfattribs or {}
+    dxfattribs = dict(dxfattribs or {})
     dxfattribs["flags"] = const.POLYLINE_3D_POLYLINE
     for path in tools.single_paths(paths):
         if len(path) > 0:
@@ -801,7 +825,7 @@ def to_lines(
     *,
     distance: float = MAX_DISTANCE,
     segments: int = MIN_SEGMENTS,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[Line]:
     """Convert the given `paths` into :class:`~ezdxf.entities.Line` entities.
 
@@ -819,7 +843,7 @@ def to_lines(
     """
     if isinstance(paths, Path):
         paths = [paths]
-    dxfattribs = dxfattribs or {}
+    dxfattribs = dict(dxfattribs or {})
     prev_vertex = None
     for path in tools.single_paths(paths):
         if len(path) == 0:
@@ -916,7 +940,7 @@ def to_splines_and_polylines(
     paths: Iterable[Path],
     *,
     g1_tol: float = G1_TOL,
-    dxfattribs: Optional[Dict] = None,
+    dxfattribs=None,
 ) -> Iterable[Union[Spline, Polyline]]:
     """Convert the given `paths` into :class:`~ezdxf.entities.Spline` and 3D
     :class:`~ezdxf.entities.Polyline` entities.
@@ -934,7 +958,7 @@ def to_splines_and_polylines(
     """
     if isinstance(paths, Path):
         paths = [paths]
-    dxfattribs = dxfattribs or {}
+    dxfattribs = dict(dxfattribs or {})
 
     for path in tools.single_paths(paths):
         for data in to_bsplines_and_vertices(path, g1_tol):
@@ -1070,12 +1094,12 @@ def to_matplotlib_path(paths: Iterable[Path], extrusion: "Vertex" = Z_AXIS):
     return MatplotlibPath(vertices, codes)
 
 
-# Interface to PyQt5.QtGui.QPainterPath
+# Interface to QtGui.QPainterPath
 
 
 def multi_path_from_qpainter_path(qpath) -> Path:
     """Returns a :class:`Path` objects from a `QPainterPath`_.
-    Returns a multi-path object if necessary. (requires PyQt5)
+    Returns a multi-path object if necessary. (requires Qt bindings)
 
     .. versionadded:: 0.17
 
@@ -1110,7 +1134,7 @@ def multi_path_from_qpainter_path(qpath) -> Path:
 
 def from_qpainter_path(qpath) -> Iterable[Path]:
     """Yields multiple :class:`Path` objects from a `QPainterPath`_.
-    (requires PyQt5)
+    (requires Qt bindings)
 
     .. versionadded:: 0.16
 
@@ -1126,12 +1150,12 @@ def from_qpainter_path(qpath) -> Iterable[Path]:
 
 
 def to_qpainter_path(paths: Iterable[Path], extrusion: "Vertex" = Z_AXIS):
-    """Convert the given `paths` into a :class:`PyQt5.QtGui.QPainterPath`
+    """Convert the given `paths` into a :class:`QtGui.QPainterPath`
     object.
     The `extrusion` vector is applied to all paths, all vertices are projected
     onto the plane normal to this extrusion vector. The default extrusion vector
     is the WCS z-axis. The :class:`QPainterPath` is a 2D object with :ref:`OCS`
-    coordinates and the z-elevation is lost. (requires PyQt5)
+    coordinates and the z-elevation is lost. (requires Qt bindings)
 
     Args:
         paths: iterable of :class:`Path` objects
@@ -1143,8 +1167,7 @@ def to_qpainter_path(paths: Iterable[Path], extrusion: "Vertex" = Z_AXIS):
     .. versionadded:: 0.16
 
     """
-    from PyQt5.QtGui import QPainterPath
-    from PyQt5.QtCore import QPointF
+    from ezdxf.addons.xqt import QPainterPath, QPointF
 
     if not Z_AXIS.isclose(extrusion):
         paths = tools.transform_paths_to_ocs(paths, OCS(extrusion))
